import polars as pl
from collections import defaultdict


class PriceTable:
    def __init__(
            self,
            df_route: pl.DataFrame,
            col_route_origin_state: str,
            col_route_destination_state: str,
            df_zone: pl.DataFrame,
            col_zone_state: str,
            col_zone_city: str,
            df_price: pl.DataFrame,
            col_price_route: str,
            col_price_zone: str,
            col_price_weight_group: str,
            col_price_basic_fee: str,
            col_price_increment: str,
            col_price_minus: str,
    ):
        # route
        self.df_route = df_route
        self.col_route_origin_state = col_route_origin_state
        self.col_route_destination_state = col_route_destination_state
        # zone
        self.df_zone = df_zone
        self.col_zone_state = col_zone_state
        self.col_zone_city = col_zone_city
        # price
        self.df_price = df_price
        self.col_price_route = col_price_route
        self.col_price_zone = col_price_zone
        self.col_price_weight_group = col_price_weight_group
        self.col_price_basic_fee = col_price_basic_fee
        self.col_price_increment = col_price_increment
        self.col_price_minus = col_price_minus

    def _preprocess(self, cols: list, name: str, df):
        df_pre_process = (
            df
            .with_columns(
                pl.concat_str([
                    pl.col(cols[0]).str.to_lowercase().str.strip_chars(),
                    pl.col(cols[1]).str.to_lowercase().str.strip_chars()
                ], separator='-').alias(name)
            )
            .drop(cols)
        )
        return df_pre_process

    def run(self):
        # route
        col_route = [self.col_route_origin_state, self.col_route_destination_state]
        self.df_route = self._preprocess(col_route, 'route_combine', self.df_route)

        # zone
        col_zone = [self.col_zone_state, self.col_zone_city]
        self.df_zone = self._preprocess(col_zone, 'zone_combine', self.df_zone)

        # price
        col_num = [
            self.col_price_route,
            self.col_price_zone,
            self.col_price_weight_group,
            self.col_price_basic_fee,
            self.col_price_increment,
            self.col_price_minus
        ]
        df_default = (
            self.df_price
            .select(col_num)
            .with_columns(
                pl.col(i).cast(pl.Float32, strict=False).fill_null(0)
                for i in col_num[2:]
            )
        )

        price = defaultdict(lambda: defaultdict(dict))
        for row in df_default.to_numpy():
            price[row[0]][row[1]].update({round(row[2], 2): row[3:].tolist()})

        return self.df_route, self.df_zone, price



from pathlib import Path
import duckdb
from src.price_mapping.utilities import PriceTable
from src.price_mapping.main import ESFLogic
from core_pro import Sheet


# data
path = Path('/Users/kevinkhang/Downloads/fe_esf/2024-05-01/raw')
file = sorted(path.glob('*available*.parquet'))

# table price
sh = '1kAE5QZD7vDCxODwRAD0ocdusiqPvXHNpUjJmvyMzZes'
df_route = Sheet(sh).google_sheet_into_df('route', 'A:C')
df_zone = Sheet(sh).google_sheet_into_df('zone', 'A:C')
df_price = Sheet(sh).google_sheet_into_df('default', 'A:G')

table = PriceTable(
    df_route=df_route,
    col_route_origin_state='Origin_State',
    col_route_destination_state='Destination_State',
    df_zone=df_zone,
    col_zone_city='City',
    col_zone_state='State',
    df_price=df_price,
    col_price_zone='Zone(O)',
    col_price_route='Route(O)',
    col_price_weight_group='Weight',
    col_price_basic_fee='*Original Fee(R)',
    col_price_increment='Increment amount(O)',
    col_price_minus='Minus'
)
df_route, df_zone, price = table.run()

# data
query = f"""
with t as (
    select * 
    , lower(buyer_address_state) || '-' || lower(seller_pickup_address) route_combine
    , lower(buyer_address_state) || '-' || lower(buyer_address_city) zone_combine
    from read_parquet('{path / f'*available*.parquet'}')
)
select t.*
,r.Route route_
,z.Zone zone_
from t
left join df_route r on t.route_combine = r.route_combine
left join df_zone z on t.zone_combine = z.zone_combine
where seller_type = 'Default'
"""
df = duckdb.sql(query).pl()
esf = ESFLogic(
    price_dict=price,
    data=df,
    col_route='route_',
    col_zone='zone_',
    col_weight='final_weight'
)
df = esf.run()



from src.price_mapping.utilities import PriceTable
from core_pro import Sheet
import polars as pl


sh = '1kAE5QZD7vDCxODwRAD0ocdusiqPvXHNpUjJmvyMzZes'
df_route = Sheet(sh).google_sheet_into_df('route', 'A:C')
df_zone = Sheet(sh).google_sheet_into_df('zone', 'A:C')
df_price = Sheet(sh).google_sheet_into_df('default', 'A:G')

table = PriceTable(
    df_route=df_route,
    col_route_origin_state='Origin_State',
    col_route_destination_state='Destination_State',
    df_zone=df_zone,
    col_zone_city='City',
    col_zone_state='State',
    df_price=df_price,
    col_price_zone='Zone(O)',
    col_price_route='Route(O)',
    col_price_weight_group='Weight',
    col_price_basic_fee='*Original Fee(R)',
    col_price_increment='Increment amount(O)',
    col_price_minus='Minus'
)
df_route, df_zone, price = table.run()



import duckdb
import polars as pl
from collections import defaultdict
import numpy as np
from tqdm import tqdm


class ESFLogic:
    def __init__(
            self,
            data,
            price_dict: defaultdict,
            col_route: str = 'route_',
            col_zone: str = 'zone_',
            col_weight: str = 'weight_',
            debug: bool = False
    ):
        self.data = data
        self.price_dict = price_dict
        self.col_route = col_route
        self.col_zone = col_zone
        self.col_weight = col_weight
        self.debug = debug

    def fee(self, array: list):
        # init
        route, urban, s_weight = array
        # lookups
        search_urban = self.price_dict.get(route, {}).get(urban)
        if not search_urban:
            return [0, 0, 0]
        # pre-calculate sorted keys for weight lookup
        sorted_keys = sorted((float(_) for _ in search_urban.keys()))
        # binary search
        weight = min(np.searchsorted(sorted_keys, s_weight), len(sorted_keys) - 1)
        weight = sorted_keys[weight]
        return search_urban[weight]

    def run(self):
        # mapping
        col = [self.col_route, self.col_zone, self.col_weight]
        fee_group_lst = [self.fee(i) for i in tqdm(self.data[col].to_numpy(), desc='Route-Zone-Weight Mapping')]
        col_name = ['basic_fee', 'extra_fee', 'weight_group']
        dict_ = {v: [_[i] for _ in fee_group_lst] for i, v in enumerate(col_name)}

        # df fee
        df_fee = pl.DataFrame(dict_)
        data = pl.concat([self.data, df_fee], how='horizontal')
        del self.data, df_fee

        # post process
        query = f"""
        select *
        , basic_fee + extra_fee * ceil(({self.col_weight} - weight_group) / 0.5) fee
        from data
        """
        data = duckdb.sql(query).pl()
        if not self.debug:
            data = data.drop(col_name)
        return data
